@@ -18,6 +18,9 @@
 
 #include <cassert>
 #include <cstring>
+#ifdef HAVE_IMLIB
+	#include <Imlib2.h>
+#endif /* HAVE_IMLIB */
 
 #include "im/im.h"
 #include "im/account.h"
@@ -69,7 +72,7 @@ string Account::getUsername() const
 string Account::getPassword() const
 {
 	assert(isValid());
-	return purple_account_get_password(account);
+	return p2s(purple_account_get_password(account));
 }
 
 void Account::setPassword(string password)
@@ -131,110 +134,260 @@ bool Account::callCommand(const string& cmd) const
 	return false;
 }
 
-vector<Protocol::Option> Account::getOptions() const
+Protocol::Options Account::getOptions() const
 {
 	assert(isValid());
 
-	vector<Protocol::Option> options = getProtocol().getOptions();
-	FOREACH(vector<Protocol::Option>, options, it)
+	Protocol::Options options = getProtocol().getOptions();
+	FOREACH(Protocol::Options, options, it)
 	{
-		Protocol::Option& option = *it;
+		Protocol::Option& option = it->second;
 		switch(option.getType())
 		{
-			case PURPLE_PREF_STRING:
+			case Protocol::Option::STR:
 				option.setValue(purple_account_get_string(account, option.getName().c_str(), option.getValue().c_str()));
 				break;
-			case PURPLE_PREF_INT:
+			case Protocol::Option::INT:
 				option.setValue(t2s(purple_account_get_int(account, option.getName().c_str(), option.getValueInt())));
 				break;
-			case PURPLE_PREF_BOOLEAN:
+			case Protocol::Option::BOOL:
 				option.setValue(purple_account_get_bool(account, option.getName().c_str(), option.getValueBool()) ? "true" : "false");
 				break;
-			default:
+			case Protocol::Option::ACCID:
+				option.setValue(getID());
+				break;
+			case Protocol::Option::STATUS_CHANNEL:
+				option.setValue(getStatusChannelName());
+				break;
+			case Protocol::Option::PASSWORD:
+				option.setValue(getPassword());
+				break;
+			case Protocol::Option::SERVER_ALIASES:
+				option.setValue(hasServerAliases() ? "true" : "false");
+				break;
+			case Protocol::Option::NONE:
 				/* not supported. */
 				break;
 		}
 	}
-	/* Minbif specific prefs */
-	options.push_back(Protocol::Option(PURPLE_PREF_STRING, "accid", "ID", getID()));
-	options.push_back(Protocol::Option(PURPLE_PREF_STRING, "status_channel", "Status Channel", getStatusChannel()));
 	return options;
 }
 
-void Account::setOptions(vector<Protocol::Option>& options)
+void Account::setOptions(const Protocol::Options& options)
 {
 	assert(isValid());
-	FOREACH(vector<Protocol::Option>, options, it)
+	for(Protocol::Options::const_iterator it = options.begin(); it != options.end(); ++it)
 	{
-		Protocol::Option& option = *it;
+		const Protocol::Option& option = it->second;
 
 		switch(option.getType())
 		{
-			case PURPLE_PREF_STRING:
-				/* XXX This is fucking ugly. */
-				if(option.getName() == "accid")
-				{
-					if(getID() == option.getValue())
-						{}
-					else if(isConnected())
-						b_log[W_ERR] << "Can't change ID of a connected account";
-					else if(Purple::getIM()->getAccount(option.getValue()).isValid() == true)
-						b_log[W_ERR] << "This ID is already used.";
-					else
-						setID(option.getValue());
-				}
-				else if(option.getName() == "status_channel")
-				{
-					if(getStatusChannel() == option.getValue())
-						{}
-					else if(!irc::Channel::isChanName(option.getValue()) || !irc::Channel::isStatusChannel(option.getValue()))
-						b_log[W_ERR] << "'" << option.getValue() << "' is not a valid channel name";
-					else
-						setStatusChannel(option.getValue());
-				}
-				else
-					purple_account_set_string(account,
-								  option.getName().c_str(),
-								  option.getValue().c_str());
+			case Protocol::Option::STR:
+				purple_account_set_string(account,
+							  option.getName().c_str(),
+							  option.getValue().c_str());
 				break;
-			case PURPLE_PREF_INT:
+			case Protocol::Option::INT:
 				purple_account_set_int(account,
 						       option.getName().c_str(),
 						       option.getValueInt());
 				break;
-			case PURPLE_PREF_BOOLEAN:
+			case Protocol::Option::BOOL:
 				purple_account_set_bool(account,
 						        option.getName().c_str(),
 							option.getValueBool());
 				break;
-			default:
-				/* not supported. */
+			case Protocol::Option::ACCID:
+			{
+				string value = option.getValue();
+				if(getID() == value)
+					break;
+
+				if(isConnected())
+					throw Protocol::OptionError("Can't change ID of a connected account");
+
+				if(value.empty())
+					value = Purple::getNewAccountName(proto, *this);
+				else if(Purple::getIM()->getAccount(value).isValid() == true)
+					throw Protocol::OptionError("This ID is already used.");
+
+				setID(value);
+				break;
+			}
+			case Protocol::Option::STATUS_CHANNEL:
+				if(getStatusChannelName() == option.getValue())
+					break;
+
+				if(!irc::Channel::isChanName(option.getValue()) || !irc::Channel::isStatusChannel(option.getValue()))
+					throw Protocol::OptionError("'" + option.getValue() + "' is not a valid channel name");
+				setStatusChannelName(option.getValue());
+				break;
+			case Protocol::Option::PASSWORD:
+				if (option.getValue() == getPassword())
+					break;
+
+				if (option.getValue().empty())
+					purple_account_set_remember_password(account, FALSE);
+				else
+				{
+					setPassword(option.getValue());
+					purple_account_set_remember_password(account, TRUE);
+				}
+				break;
+			case Protocol::Option::SERVER_ALIASES:
+				setServerAliases(option.getValueBool());
+				break;
+			case Protocol::Option::NONE:
+				/* Not happy. */
 				break;
 		}
 	}
-
 }
 
-void Account::setBuddyIcon(const string& filename)
+void Account::setBuddyIcon(string filename)
 {
 	assert(isValid());
 	PurplePlugin *plug = purple_find_prpl(purple_account_get_protocol_id(account));
 	if (plug) {
 		PurplePluginProtocolInfo *prplinfo = PURPLE_PLUGIN_PROTOCOL_INFO(plug);
 		if (prplinfo != NULL &&
-		    purple_account_get_bool(account, "use-global-buddyicon", TRUE) &&
-		    prplinfo->icon_spec.format) {
+		    //purple_account_get_bool(account, "use-global-buddyicon", TRUE) &&
+		    prplinfo->icon_spec.format)
+		{
+			GError* temp_error = NULL;
+#ifdef HAVE_IMLIB
+			Imlib_Image img = imlib_load_image(filename.c_str());
+			char* temp_filename = NULL;
+
+			if (img) {
+				int temp_fd;
+
+				/* Create a stupid tmp file, write it, close it. Save image as png in it. Fuck it. */
+				temp_fd = g_file_open_tmp("minbif_new_icon_XXXXXX", &temp_filename, &temp_error);
+				if (temp_error) {
+					b_log[W_ERR] << "Unable to create a temporary file: " << temp_error->message;
+					g_error_free(temp_error);
+					temp_filename = NULL;
+				}
+				else
+				{
+					char** prpl_formats = g_strsplit(prplinfo->icon_spec.format,",",0);
+					ImlibLoadError err = IMLIB_LOAD_ERROR_UNKNOWN;
+
+					close(temp_fd);
+					/* Try to encode in a supported format. */
+					imlib_context_set_image(img);
+					for (size_t i = 0; prpl_formats[i] && err != IMLIB_LOAD_ERROR_NONE; ++i)
+					{
+						imlib_image_set_format(prpl_formats[i]);
+						imlib_save_image_with_error_return(temp_filename, &err);
+					}
+
+					if (err != IMLIB_LOAD_ERROR_NONE)
+						b_log[W_ERR] << "Unable to encode image for " << getID();
+					else
+						filename = temp_filename;
+					g_strfreev(prpl_formats);
+				}
+				imlib_free_image();
+			}
+
+#endif /* HAVE_IMLIB */
 			gchar *contents;
 			gsize len;
-			if (g_file_get_contents(filename.c_str(), &contents, &len, NULL))
-				purple_buddy_icons_set_account_icon(account, (guchar *)contents, len);
-
-			purple_account_set_buddy_icon_path(account, filename.c_str());
+			if (g_file_get_contents(filename.c_str(), &contents, &len, &temp_error))
+			{
+				if ((prplinfo->icon_spec.max_filesize != 0) &&
+				    (len > prplinfo->icon_spec.max_filesize))
+				{
+					b_log[W_ERR] << "Unable to set image on " << getID() << ": image too large";
+				}
+				else
+				{
+					purple_buddy_icons_set_account_icon(account, (guchar *)contents, len);
+					/* Useless, and a temp filename is not safe:
+					 *   purple_account_set_buddy_icon_path(account, temp_filename);
+					 */
+				}
+			}
+			else if (temp_error)
+			{
+				b_log[W_ERR] << "Unable to get image content on " << getID() << ": " << temp_error->message;
+				g_error_free(temp_error);
+			}
+#ifdef HAVE_IMLIB
+			if (temp_filename) {
+				unlink(temp_filename);
+				g_free(temp_filename);
+			}
+#endif /* HAVE_IMLIB */
 		}
 	}
 
 }
 
+void Account::setStatus(PurpleStatusPrimitive prim, string message)
+{
+	assert(isValid());
+
+	const char *id;
+	PurpleStatus* status;
+
+	if (prim == PURPLE_STATUS_UNSET)
+	{
+		status = purple_account_get_active_status(account);
+		id = purple_status_get_id(status);
+	}
+	else
+	{
+		id = purple_primitive_get_id_from_type(prim);
+		status = purple_account_get_status(account, id);
+	}
+
+	if (!status)
+		return;
+
+	irc::StatusChannel* chan = getStatusChannel();
+	if (message.empty() && chan)
+		message = chan->getTopic();
+	if (message.empty() && prim != PURPLE_STATUS_UNSET)
+	{
+		/* Get the status' message only if we change it. */
+		message = p2s(purple_status_get_attr_string(status, "message"));
+		if (!message.empty() && chan && purple_primitive_get_type_from_id(id) == PURPLE_STATUS_AVAILABLE)
+			chan->setTopic(NULL, message);
+	}
+
+	purple_account_set_status(account, id, TRUE, "message", message.c_str(), (char*)NULL);
+}
+
+PurpleStatusPrimitive Account::getStatus() const
+{
+	assert(isValid());
+
+	PurpleStatus *s = purple_account_get_active_status(account);
+	if (!s)
+		return PURPLE_STATUS_UNSET;
+
+	return purple_primitive_get_type_from_id(purple_status_get_id(s));
+}
+
+string Account::getStatusMessage(PurpleStatusPrimitive prim) const
+{
+	assert(isValid());
+
+	PurpleStatus *s;
+	if (prim == PURPLE_STATUS_UNSET)
+		s = purple_account_get_active_status(account);
+	else
+		s = purple_account_get_status(account, purple_primitive_get_id_from_type(prim));
+
+	if (!s)
+		return "";
+
+	return p2s(purple_status_get_attr_string(s, "message"));
+}
+
 void Account::setID(string id) const
 {
 	assert(isValid());
@@ -247,20 +400,32 @@ string Account::getID() const
 	string n = purple_account_get_ui_string(account, MINBIF_VERSION_NAME, "id", "");
 	if(n.empty())
 	{
-		n = Purple::getNewAccountName(proto);
+		n = Purple::getNewAccountName(proto, *this);
 		setID(n);
 	}
 	return n;
 }
 
-string Account::getStatusChannel() const
+bool Account::hasServerAliases() const
+{
+	assert(isValid());
+	return purple_account_get_ui_bool(account, MINBIF_VERSION_NAME, "server_aliases", true);
+}
+
+void Account::setServerAliases(bool b)
+{
+	assert(isValid());
+	purple_account_set_ui_bool(account, MINBIF_VERSION_NAME, "server_aliases", b);
+}
+
+string Account::getStatusChannelName() const
 {
 	assert(isValid());
 	string n = purple_account_get_ui_string(account, MINBIF_VERSION_NAME, "channel", "");
 	return n;
 }
 
-void Account::setStatusChannel(const string& c)
+void Account::setStatusChannelName(const string& c)
 {
 	assert(isValid());
 
@@ -353,6 +518,14 @@ vector<Buddy> Account::getBuddies() const
 	return buddies;
 }
 
+void Account::updatedAllBuddies() const
+{
+	assert(isValid());
+	vector<Buddy> buddies = getBuddies();
+	for(vector<Buddy>::iterator it = buddies.begin(); it != buddies.end(); ++it)
+		it->updated();
+}
+
 void Account::displayRoomList() const
 {
 	assert(isValid());
@@ -403,13 +576,32 @@ void Account::removeReconnection(bool verbose) const
 	purple_account_set_ui_int(account, MINBIF_VERSION_NAME, "id-reconnect", -1);
 }
 
-void Account::createStatusChannel() const
+irc::StatusChannel* Account::getStatusChannel() const
+{
+	irc::IRC* irc = Purple::getIM()->getIRC();
+	irc::StatusChannel* chan;
+	string channame = getStatusChannelName();
+
+	if(!irc::Channel::isStatusChannel(channame))
+		return NULL;
+
+	chan = dynamic_cast<irc::StatusChannel*>(irc->getChannel(channame));
+	if(!chan)
+	{
+		b_log[W_ERR] << "Status channel " << channame << " is not found.";
+		return NULL;
+	}
+
+	return chan;
+}
+
+void Account::createStatusChannel()
 {
 	assert(isValid());
 
 	irc::IRC* irc = Purple::getIM()->getIRC();
 	irc::StatusChannel* chan;
-	string channame = getStatusChannel();
+	string channame = getStatusChannelName();
 
 	if(!irc::Channel::isStatusChannel(channame))
 		return;
@@ -421,6 +613,16 @@ void Account::createStatusChannel() const
 		irc->addChannel(chan);
 		irc->getUser()->join(chan, irc::ChanUser::OP);
 	}
+
+	string status = getStatusMessage(PURPLE_STATUS_AVAILABLE);
+	if (chan->getTopic() != status)
+	{
+		if (!status.empty())
+			chan->setTopic(NULL, status);
+		else
+			setStatus(PURPLE_STATUS_UNSET, chan->getTopic());
+	}
+
 	chan->addAccount(*this);
 
 	vector<Buddy> buddies = getBuddies();
@@ -428,23 +630,15 @@ void Account::createStatusChannel() const
 		b->updated();
 }
 
-void Account::leaveStatusChannel() const
+void Account::leaveStatusChannel()
 {
 	assert(isValid());
 
 	irc::IRC* irc = Purple::getIM()->getIRC();
-	irc::StatusChannel* chan;
-	string channame = getStatusChannel();
-
-	if(!irc::Channel::isStatusChannel(channame))
+	irc::StatusChannel* chan = getStatusChannel();
+	if (!chan)
 		return;
 
-	chan = dynamic_cast<irc::StatusChannel*>(irc->getChannel(channame));
-	if(!chan)
-	{
-		b_log[W_ERR] << "Status channel " << channame << " is not found.";
-		return;
-	}
 	chan->removeAccount(*this);
 
 	vector<Buddy> buddies = getBuddies();
@@ -453,7 +647,7 @@ void Account::leaveStatusChannel() const
 			b->getNick()->part(chan, "Leaving status channel");
 
 	if(chan->countAccounts() == 0)
-		irc->removeChannel(channame);
+		irc->removeChannel(chan->getName());
 }
 
 vector<string> Account::getDenyList() const
@@ -467,18 +661,25 @@ vector<string> Account::getDenyList() const
 	return list;
 }
 
-void Account::deny(const string& who) const
+bool Account::deny(const string& who) const
 {
 	assert(isValid());
 
-	purple_privacy_deny(account, who.c_str(), FALSE, FALSE);
+	if (account->perm_deny != PURPLE_PRIVACY_DENY_USERS)
+	{
+		account->perm_deny = PURPLE_PRIVACY_DENY_USERS;
+		if (purple_account_is_connected(account))
+			serv_set_permit_deny(purple_account_get_connection(account));
+	}
+
+	return purple_privacy_deny_add(account, who.c_str(), FALSE);
 }
 
-void Account::allow(const string& who) const
+bool Account::allow(const string& who) const
 {
 	assert(isValid());
 
-	purple_privacy_allow(account, who.c_str(), FALSE, FALSE);
+	return purple_privacy_deny_remove(account, who.c_str(), FALSE);
 }
 
 void Account::addBuddy(const string& username, const string& group) const
@@ -521,6 +722,7 @@ bool Account::supportsChats() const
 map<string, string> Account::getChatParameters() const
 {
 	assert(isValid());
+	assert(isConnected());
 
 	PurpleConnection* gc = purple_account_get_connection(account);
 	map<string, string> m;
@@ -981,7 +1183,6 @@ void Account::connected(PurpleConnection* gc)
 	b_log[W_INFO|W_SNO] << "Connection to " << account.getServername() << " established!";
 	irc->addServer(new irc::RemoteServer(irc, account));
 	account.flushChannelJoins();
-
 }
 
 void Account::account_signed_on_cb(PurpleConnection *gc, gpointer event)
@@ -1050,9 +1251,17 @@ void Account::disconnect_reason(PurpleConnection *gc,
 				const char *text)
 {
 	Account acc(gc->account);
-	int delay = acc.delayReconnect();
 	b_log[W_ERR|W_SNO] << "Error(" << acc.getServername() << "): " << text;
-	b_log[W_ERR|W_SNO] << "Reconnection in " << delay << " seconds";
+	switch(reason)
+	{
+	case PURPLE_CONNECTION_ERROR_NETWORK_ERROR:
+	case PURPLE_CONNECTION_ERROR_ENCRYPTION_ERROR:
+	case PURPLE_CONNECTION_ERROR_OTHER_ERROR:
+		b_log[W_ERR|W_SNO] << "Reconnection in " << acc.delayReconnect() << " seconds";
+		break;
+	default:
+		break; /* Do not auto reconnect. */
+	}
 }
 
 }; /* namespace im */
