@@ -1,6 +1,6 @@
 /*
  * Minbif - IRC instant messaging gateway
- * Copyright(C) 2010 Marc Dequènes (Duck)
+ * Copyright(C) 2010 Romain Bignon, Marc Dequènes (Duck)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -37,10 +37,11 @@ namespace im
 	class AuthPAM : public Auth
 	{
 	public:
-		AuthPAM(irc::IRC* _irc, string _username);
+		AuthPAM(irc::IRC* _irc, const string& _username);
 		~AuthPAM();
 		bool exists();
-		bool authenticate(const string password);
+		bool authenticate(const string& password);
+		im::IM* create(const string& password);
 		bool setPassword(const string& password);
 		string getPassword() const;
 
@@ -50,6 +51,7 @@ namespace im
 		struct _pam_conv_func_data pam_conv_func_data;
 
 		void close(int retval = PAM_SUCCESS);
+		bool checkPassword(const string& password);
 	};
 };
 
