@@ -199,12 +199,12 @@ PurpleConversationType Conversation::getType() const
 	return purple_conversation_get_type(conv);
 }
 
-void Conversation::setNick(irc::Nick* n, bool purge_unknown)
+void Conversation::setNick(irc::ConvNick* n, bool purge_unknown)
 {
 	assert(isValid());
 	assert(getType() == PURPLE_CONV_TYPE_IM);
 
-	irc::Nick* last = getNick();
+	irc::ConvNick* last = getNick();
 	conv->ui_data = n;
 
 	if(last && last != n)
@@ -214,14 +214,16 @@ void Conversation::setNick(irc::Nick* n, bool purge_unknown)
 			irc::IRC* irc = Purple::getIM()->getIRC();
 			irc->removeNick(last->getNickname());
 		}
+		else
+			last->setConversation(Conversation());
 	}
 }
 
-irc::Nick* Conversation::getNick() const
+irc::ConvNick* Conversation::getNick() const
 {
 	assert(isValid());
 	assert(getType() == PURPLE_CONV_TYPE_IM);
-	return static_cast<irc::Nick*>(conv->ui_data);
+	return static_cast<irc::ConvNick*>(conv->ui_data);
 }
 
 void Conversation::setChannel(irc::ConversationChannel* c) const
@@ -323,6 +325,7 @@ void Conversation::destroyChannel() const
 
 	irc::IRC* irc = Purple::getIM()->getIRC();
 	irc->removeChannel(getChannel()->getName());
+	setChannel(NULL);
 }
 
 PurpleConvChat* Conversation::getPurpleChat() const
@@ -421,19 +424,22 @@ void Conversation::recvMessage(string from, string text, bool action)
 	{
 		case PURPLE_CONV_TYPE_IM:
 		{
-			irc::Nick* n = getNick();
+			irc::ConvNick* n = getNick();
 			if(!n)
 			{
 				/* There isn't any Nick associated to this conversation. Try to
 				 * find a Nick with this conversation object. */
-				n = irc->getNick(*this);
+				n = dynamic_cast<irc::ConvNick*>(irc->getNick(*this));
 				if(!n)
 				{
 					/* If there isn't any conversation, try to find a buddy. */
-					n = irc->getNick(from);
+					n = dynamic_cast<irc::Buddy*>(irc->getNick(from));
 
-					if(!n || !dynamic_cast<irc::Buddy*>(n))
+					if(!n)
 					{
+						if (!Purple::getIM()->hasAcceptNoBuddiesMessages())
+							return;
+
 						/* Ok, there isn't any buddy, so I create an unknown buddy to chat with him. */
 						n = new irc::UnknownBuddy(irc->getServer(getAccount().getServername()), *this);
 						while((irc->getNick(n->getNickname())))
@@ -442,17 +448,12 @@ void Conversation::recvMessage(string from, string text, bool action)
 					}
 				}
 				setNick(n);
+				n->setConversation(*this);
 			}
 
 			string line;
 			while((line = stringtok(text, "\n\r")).empty() == false)
-			{
-				if(action)
-					line = "\001ACTION " + line + "\001";
-				irc->getUser()->send(irc::Message(MSG_PRIVMSG).setSender(n)
-									      .setReceiver(irc->getUser())
-									      .addArg(line));
-			}
+				n->sendMessage(irc->getUser(), line, action);
 			break;
 		}
 		case PURPLE_CONV_TYPE_CHAT:
@@ -579,15 +580,12 @@ void Conversation::destroy(PurpleConversation* c)
 	{
 		case PURPLE_CONV_TYPE_IM:
 		{
+			conv.setNick(NULL);
+
 			irc::IRC* irc = Purple::getIM()->getIRC();
 			irc::ConvNick* n = dynamic_cast<irc::ConvNick*>(irc->getNick(conv));
 			if(n)
-			{
-				conv.setNick(NULL);
-				if(dynamic_cast<irc::UnknownBuddy*>(n))
-					irc->removeNick(n->getNickname());
 				n->setConversation(Conversation());
-			}
 			break;
 		}
 		case PURPLE_CONV_TYPE_CHAT:
@@ -792,16 +790,19 @@ void Conversation::buddy_typing(PurpleAccount* account, const char* who, gpointe
 			irc->getUser()->send(irc::Message(MSG_PRIVMSG).setSender(n)
 						.setReceiver(irc->getUser())
 						.addArg("\1TYPING 1\1"));
+			break;
 
 		case PURPLE_TYPED:
 			irc->getUser()->send(irc::Message(MSG_PRIVMSG).setSender(n)
 						.setReceiver(irc->getUser())
 						.addArg("\1TYPING 2\1"));
+			break;
 
 		case PURPLE_NOT_TYPING:
 			irc->getUser()->send(irc::Message(MSG_PRIVMSG).setSender(n)
 						.setReceiver(irc->getUser())
 						.addArg("\1TYPING 0\1"));
+			break;
 	}
 }
 
