@@ -25,6 +25,7 @@
 #include "purple.h"
 #include "irc/irc.h"
 #include "irc/user.h"
+#include "core/caca_image.h"
 #include "core/util.h"
 #include "core/log.h"
 #include "core/config.h"
@@ -118,9 +119,11 @@ public:
 	{}
 };
 
-Request::Request(PurpleRequestType _type, const string& _title, const string& _question)
-	: title(_title),
-	  question(_question),
+Request::Request(PurpleRequestType _type, const Account& _account, const string& _title, const string& _primary, const string& _secondary)
+	: account(_account),
+	  title(_title),
+	  primary(_primary),
+	  secondary(_secondary),
 	  type(_type)
 {
 
@@ -131,6 +134,9 @@ RequestFieldList::~RequestFieldList()
 	map<string, RequestField*>::iterator it;
 	for(it = fields.begin(); it != fields.end(); ++it)
 		delete it->second;
+
+	for(vector<Request*>::iterator r = subrequests.begin(); r != subrequests.end(); r = subrequests.erase(r))
+		delete *r;
 }
 
 void RequestFieldList::addField(RequestField* field)
@@ -150,17 +156,51 @@ RequestField* RequestFieldList::getField(const string& label) const
 	return it->second;
 }
 
+void RequestFieldList::addSubrequest(Request* request)
+{
+	subrequests.push_back(request);
+}
+
 void RequestFieldList::process(const string& answer) const
 {
-	RequestField* field = getField(answer);
-	field->runCallback();
+	if (!subrequests.empty())
+		subrequests.front()->process(answer);
+	else
+	{
+		RequestField* field = getField(answer);
+		field->runCallback();
+	}
+}
+
+void RequestFieldList::close()
+{
+	if (!subrequests.empty())
+	{
+		delete subrequests.front();
+		subrequests.erase(subrequests.begin());
+		display();
+	}
+	else
+		Request::close();
 }
 
 void RequestFieldList::display() const
 {
+	if (!subrequests.empty())
+	{
+		subrequests.front()->display();
+		return;
+	}
+
 	irc::IRC* irc = Purple::getIM()->getIRC();
-	nick->privmsg(irc->getUser(), "New request: " + title);
-	nick->privmsg(irc->getUser(), question);
+	if(account.isValid())
+		nick->privmsg(irc->getUser(), "[" + account.getID() + "] New request: " + title);
+	else
+		nick->privmsg(irc->getUser(), "New request: " + title);
+	if(primary.empty() == false)
+		nick->privmsg(irc->getUser(), primary);
+	if(secondary.empty() == false)
+		nick->privmsg(irc->getUser(), secondary);
 	for(map<string, RequestField*>::const_iterator it = fields.begin();
 	    it != fields.end();
 	    ++it)
@@ -177,8 +217,15 @@ void RequestInput::process(const string& answer) const
 void RequestInput::display() const
 {
 	irc::IRC* irc = Purple::getIM()->getIRC();
-	nick->privmsg(irc->getUser(), "New request: " + title);
-	nick->privmsg(irc->getUser(), question);
+	if(account.isValid())
+		nick->privmsg(irc->getUser(), "[" + account.getID() + "] New request: " + title);
+	else
+		nick->privmsg(irc->getUser(), "New request: " + title);
+	if(primary.empty() == false)
+		nick->privmsg(irc->getUser(), primary);
+	if(secondary.empty() == false)
+		nick->privmsg(irc->getUser(), secondary);
+
 	nick->privmsg(irc->getUser(), "Default value is: " + default_value);
 	nick->privmsg(irc->getUser(), "Type answer:");
 }
@@ -248,11 +295,13 @@ Request* Request::getFirstRequest()
 	return requests.empty() ? NULL : requests.front();
 }
 
-void Request::addRequest(Request* request)
+Request* Request::addRequest(Request* request)
 {
 	requests.push_back(request);
 	if(requests.size() == 1)
 		request->display();
+
+	return request;
 }
 
 void* Request::notify_message(PurpleNotifyMsgType type, const char *title,
@@ -372,8 +421,8 @@ void Request::closeRequest(const Request* request)
 			continue;
 
 		bool first = (it == requests.begin());
-		delete *it;
 		requests.erase(it);
+		delete request;
 
 		if(first)
 		{
@@ -398,10 +447,9 @@ void* Request::request_input(const char *title, const char *primary,
 			PurpleAccount *account, const char *who, PurpleConversation *conv,
 			void *user_data)
 {
-	RequestInput* request = new RequestInput(PURPLE_REQUEST_INPUT, title ? title : "", primary ? primary : "", default_value ? default_value : "", (PurpleRequestInputCb)ok_cb, user_data);
-	addRequest(request);
-
-	return requests.back();
+	RequestInput* request = new RequestInput(PURPLE_REQUEST_INPUT, Account(account), title ? title : "", primary ? primary : "", secondary ? secondary : "",
+	                                         default_value ? default_value : "", (PurpleRequestInputCb)ok_cb, user_data);
+	return addRequest(request);
 }
 
 void* Request::request_action(const char *title, const char *primary,
@@ -410,7 +458,7 @@ void* Request::request_action(const char *title, const char *primary,
 			void *user_data, size_t actioncount,
 			va_list actions)
 {
-	RequestFieldList* request = new RequestFieldList(PURPLE_REQUEST_ACTION, title ? title : "", primary ? primary : "");
+	RequestFieldList* request = new RequestFieldList(PURPLE_REQUEST_ACTION, Account(account), title ? title : "", primary ? primary : "", secondary ? secondary : "");
 	for(size_t i = 0; i < actioncount; ++i)
 	{
 		const char *text = va_arg(actions, const char *);
@@ -419,9 +467,7 @@ void* Request::request_action(const char *title, const char *primary,
 
 		request->addField(new RequestFieldAction<PurpleRequestActionCb>((int)i, strlower(stringtok(tmp, "_ ")), text, callback, user_data));
 	}
-	addRequest(request);
-
-	return requests.back();
+	return addRequest(request);
 }
 
 void* Request::request_choice(const char *title, const char *primary,
@@ -431,7 +477,7 @@ void* Request::request_choice(const char *title, const char *primary,
 			PurpleAccount *account, const char *who, PurpleConversation *conv,
 			void *user_data, va_list choices)
 {
-	RequestFieldList* request = new RequestFieldList(PURPLE_REQUEST_CHOICE, title ? title : "", primary ? primary : "");
+	RequestFieldList* request = new RequestFieldList(PURPLE_REQUEST_CHOICE, Account(account), title ? title : "", primary ? primary : "", secondary ? secondary : "");
 	const char* text;
 
 	while ((text = va_arg(choices, const char *)))
@@ -443,9 +489,7 @@ void* Request::request_choice(const char *title, const char *primary,
 	}
 	request->addField(new RequestFieldAction<PurpleRequestChoiceCb>(0, "cancel", "Cancel", (PurpleRequestChoiceCb)cancel_cb, user_data));
 
-	addRequest(request);
-
-	return requests.back();
+	return addRequest(request);
 }
 
 class RequestFieldsString : public RequestInput
@@ -453,8 +497,8 @@ class RequestFieldsString : public RequestInput
 	PurpleRequestField* field;
 public:
 
-	RequestFieldsString(PurpleRequestField* _field, const string& title, const string& question, const string& _default_value)
-		: RequestInput(PURPLE_REQUEST_FIELDS, title, question, _default_value, NULL, NULL),
+	RequestFieldsString(PurpleRequestField* _field, const Account& account, const string& title, const string& primary, const string& secondary, const string& _default_value)
+		: RequestInput(PURPLE_REQUEST_FIELDS, account, title, primary, secondary, _default_value, NULL, NULL),
 		  field(_field)
 	{}
 
@@ -473,8 +517,8 @@ class RequestFieldsInteger : public RequestInput
 	PurpleRequestField* field;
 public:
 
-	RequestFieldsInteger(PurpleRequestField* _field, const string& title, const string& question, const string& _default_value)
-		: RequestInput(PURPLE_REQUEST_FIELDS, title, question, _default_value, NULL, NULL),
+	RequestFieldsInteger(PurpleRequestField* _field, const Account& account, const string& title, const string& primary, const string& secondary, const string& _default_value)
+		: RequestInput(PURPLE_REQUEST_FIELDS, account, title, primary, secondary, _default_value, NULL, NULL),
 		  field(_field)
 	{}
 
@@ -488,11 +532,68 @@ public:
 	}
 };
 
+class RequestFieldsImage : public Request
+{
+	PurpleRequestField* field;
+public:
+	RequestFieldsImage(const Account& account, PurpleRequestField* _field)
+		: Request(PURPLE_REQUEST_FIELDS, account, "","",""),
+		  field(_field)
+	{}
+
+	void process(const string& answer) const
+	{
+	}
+
+	void close()
+	{
+		Request::closeRequest(this);
+	}
+
+	void display() const
+	{
+		GError* temp_error = NULL;
+		gchar* temp_filename = NULL;
+		int temp_fd = g_file_open_tmp("minbif_request_XXXXXX.img", &temp_filename, &temp_error);
+
+		if(temp_error)
+		{
+			b_log[W_ERR] << "Unable to create a temporary file: " << temp_error->message;
+			g_error_free(temp_error);
+		}
+		else
+		{
+			write(temp_fd,
+			      purple_request_field_image_get_buffer(field),
+			      purple_request_field_image_get_size(field));
+			::close(temp_fd);
+
+			irc::IRC* irc = Purple::getIM()->getIRC();
+			nick->privmsg(irc->getUser(), string("Image (") + temp_filename + ")");
+
+#ifdef HAVE_CACA
+			CacaImage img(temp_filename);
+			string line, buf = img.getIRCBuffer(60);
+			while((line = stringtok(buf, "\r\n")).empty() == false)
+			{
+				nick->privmsg(irc->getUser(), line);
+			}
+#endif /* HAVE_CACA */
+		}
+
+		Request* request = Request::getFirstRequest();
+		if (request)
+			request->close();
+	}
+
+
+};
+
 class RequestFieldsList : public RequestFieldList
 {
 public:
-	RequestFieldsList(const string& title, const string& question)
-		: RequestFieldList(PURPLE_REQUEST_FIELDS, title, question)
+	RequestFieldsList(const Account& account, const string& title, const string& primary, const string& secondary)
+		: RequestFieldList(PURPLE_REQUEST_FIELDS, account, title, primary, secondary)
 	{}
 
 	virtual void close() {
@@ -578,6 +679,10 @@ void* Request::request_fields(const char *title, const char *primary,
 {
 	GList *gl, *fl;
 
+	RequestFieldList* mainrequest = new RequestFieldList(PURPLE_REQUEST_FIELDS, Account(account), title ? title : primary ? primary : "",
+	                                                     title ? (primary ? primary : (secondary ? secondary : "")) : secondary ? secondary : "",
+							     "Do you confirm?");
+
 	for(gl = purple_request_fields_get_groups(fields);
 	    gl != NULL;
 	    gl = gl->next)
@@ -598,16 +703,28 @@ void* Request::request_fields(const char *title, const char *primary,
 			if(type == PURPLE_REQUEST_FIELD_STRING)
 			{
 				const char* defval = purple_request_field_string_get_default_value(field);
-				addRequest(new RequestFieldsString(field, title ? title : primary ? primary : "", field_label ? field_label : "", defval ? defval : ""));
+				RequestFieldsString* request = new RequestFieldsString(field, Account(account),
+						                                       title ? title : primary ? primary : "", secondary ? secondary : "",
+				                                                       field_label ? field_label : "", defval ? defval : "");
+				mainrequest->addSubrequest(request);
 			}
 			else if(type == PURPLE_REQUEST_FIELD_INTEGER)
 			{
 				string defval = t2s(purple_request_field_int_get_default_value(field));
-				addRequest(new RequestFieldsInteger(field, title ? title : primary ? primary : "", field_label ? field_label : "", defval));
+				RequestFieldsInteger* request = new RequestFieldsInteger(field, Account(account),
+						                                         title ? title : primary ? primary : "", secondary ? secondary : "",
+				                                                         field_label ? field_label : "", defval);
+				mainrequest->addSubrequest(request);
+			}
+			else if(type == PURPLE_REQUEST_FIELD_IMAGE)
+			{
+				RequestFieldsImage* request = new RequestFieldsImage(Account(account), field);
+				mainrequest->addSubrequest(request);
 			}
 			else
 			{
-				RequestFieldsList* request = new RequestFieldsList(title ? title : primary ? primary : "", field_label ? field_label : "");
+				RequestFieldsList* request = new RequestFieldsList(Account(account), title ? title : primary ? primary : "",
+				                                                   secondary ? secondary : "", field_label ? field_label : "");
 				switch(type)
 				{
 					case PURPLE_REQUEST_FIELD_BOOLEAN:
@@ -632,21 +749,19 @@ void* Request::request_fields(const char *title, const char *primary,
 						break;
 					}
 					default:
-						 delete request;
-						 continue;
+						b_log[W_ERR] << "Skipped a field " << type;
+						delete request;
+						continue;
 				}
-				addRequest(request);
+				mainrequest->addSubrequest(request);
 			}
 		}
 	}
 
-	RequestFieldList* request = new RequestFieldList(PURPLE_REQUEST_FIELDS, title ? title : primary ? primary : "", "Do you confirm?");
-	request->addField(new RequestFieldTmplt<PurpleRequestFieldsCb,PurpleRequestFields*>(0, "ok", ok, (PurpleRequestFieldsCb)ok_cb, userdata, fields));
-	request->addField(new RequestFieldTmplt<PurpleRequestFieldsCb,PurpleRequestFields*>(1, "cancel", cancel, (PurpleRequestFieldsCb)cancel_cb, userdata, fields));
+	mainrequest->addField(new RequestFieldTmplt<PurpleRequestFieldsCb,PurpleRequestFields*>(0, "ok", ok, (PurpleRequestFieldsCb)ok_cb, userdata, fields));
+	mainrequest->addField(new RequestFieldTmplt<PurpleRequestFieldsCb,PurpleRequestFields*>(1, "cancel", cancel, (PurpleRequestFieldsCb)cancel_cb, userdata, fields));
 
-	addRequest(request);
-
-	return requests.back();
+	return addRequest(mainrequest);
 }
 
 void* Request::request_file(const char *title, const char *filename,
