@@ -17,18 +17,16 @@
  */
 
 #include <sys/socket.h>
-#include <fcntl.h>
-#include <netdb.h>
 #include <cstring>
 #include <algorithm>
 #include <cassert>
 #include <fstream>
+#include <fnmatch.h>
 
 #include "core/log.h"
 #include "core/util.h"
 #include "core/callback.h"
 #include "core/version.h"
-#include "core/sock.h"
 #include "core/config.h"
 #include "im/im.h"
 #include "server_poll/poll.h"
@@ -57,6 +55,7 @@ IRC::command_t IRC::commands[] = {
 	{ MSG_PING,    &IRC::m_ping,    0, 0, Nick::REGISTERED },
 	{ MSG_PONG,    &IRC::m_pong,    1, 0, Nick::REGISTERED },
 	{ MSG_VERSION, &IRC::m_version, 0, 0, Nick::REGISTERED },
+	{ MSG_INFO,    &IRC::m_info,    0, 0, Nick::REGISTERED },
 	{ MSG_WHO,     &IRC::m_who,     0, 0, Nick::REGISTERED },
 	{ MSG_WHOIS,   &IRC::m_whois,   1, 0, Nick::REGISTERED },
 	{ MSG_WHOWAS,  &IRC::m_whowas,  1, 0, Nick::REGISTERED },
@@ -84,60 +83,38 @@ IRC::command_t IRC::commands[] = {
 	{ MSG_DIE,     &IRC::m_die,     1, 0, Nick::OPER },
 };
 
-IRC::IRC(ServerPoll* _poll, int _fd, string _hostname, unsigned _ping_freq)
+IRC::IRC(ServerPoll* _poll, sock::SockWrapper* _sockw, string _hostname, unsigned _ping_freq)
 	: Server("localhost.localdomain", MINBIF_VERSION),
 	  poll(_poll),
-	  fd(_fd),
-	  read_id(-1),
+	  sockw(_sockw),
 	  read_cb(NULL),
 	  ping_id(-1),
 	  ping_freq(_ping_freq),
+	  uptime(time(NULL)),
 	  ping_cb(NULL),
 	  user(NULL),
-	  im(NULL)
+	  im(NULL),
+	  im_auth(NULL)
 {
-	struct sockaddr_storage sock;
-	socklen_t socklen = sizeof(sock);
-
-	/* Get the user's hostname. */
-	string userhost = "localhost.localdomain";
-	if(getpeername(fd, (struct sockaddr*) &sock, &socklen) == 0)
-	{
-		char buf[NI_MAXHOST+1];
-
-		if(getnameinfo((struct sockaddr *)&sock, socklen, buf, NI_MAXHOST, NULL, 0, 0) == 0)
-			userhost = buf;
-	}
-
 	/* Get my own hostname (if not given in arguments) */
 	if(_hostname.empty() || _hostname == " ")
-	{
-		if(getsockname(fd, (struct sockaddr*) &sock, &socklen) == 0)
-		{
-			char buf[NI_MAXHOST+1];
-
-			if(getnameinfo((struct sockaddr *) &sock, socklen, buf, NI_MAXHOST, NULL, 0, 0 ) == 0)
-				setName(buf);
-		}
-	}
+		setName(sockw->GetServerHostname());
 	else if(_hostname.find(" ") != string::npos)
 	{
 		/* An hostname can't contain any space. */
 		b_log[W_ERR] << "'" << _hostname << "' is not a valid server hostname";
-		throw AuthError();
+		throw sock::SockError("Wrong server hostname");
 	}
 	else
 		setName(_hostname);
 
 	/* create a callback on the sock. */
 	read_cb = new CallBack<IRC>(this, &IRC::readIO);
-	/* XXX it appears that it is not free'd */
-	read_id = glib_input_add(fd, (PurpleInputCondition)PURPLE_INPUT_READ, g_callback_input, read_cb);
+	sockw->AttachCallback(PURPLE_INPUT_READ, read_cb);
 
 	/* Create main objects and root joins command channel. */
-	user = new User(fd, this, "*", "", userhost);
+	user = new User(sockw, this, "*", "", sockw->GetClientHostname());
 	addNick(user);
-	im_auth = NULL;
 
 	/* Ping callback */
 	if(ping_freq > 0)
@@ -157,14 +134,12 @@ IRC::~IRC()
 	if (im_auth)
 		delete im_auth;
 
-	if(read_id >= 0)
-		g_source_remove(read_id);
 	if(ping_id >= 0)
 		g_source_remove(ping_id);
-	delete read_cb;
 	delete ping_cb;
-	if(fd >= 0)
-		close(fd);
+	if(sockw)
+		delete sockw;
+	delete read_cb;
 	cleanUpNicks();
 	cleanUpServers();
 	cleanUpChannels();
@@ -252,6 +227,7 @@ void IRC::addNick(Nick* nick)
 	if(users.find(nick->getNickname()) != users.end())
 		b_log[W_DESYNCH] << "/!\\ User " << nick->getNickname() << " already exists!";
 	users[nick->getNickname()] = nick;
+	nick->getServer()->addNick(nick);
 }
 
 void IRC::renameNick(Nick* nick, string newnick)
@@ -275,7 +251,7 @@ Nick* IRC::getNick(string nickname, bool case_sensitive) const
 	return it->second;
 }
 
-Nick* IRC::getNick(const im::Buddy& buddy) const
+Buddy* IRC::getNick(const im::Buddy& buddy) const
 {
 	map<string, Nick*>::const_iterator it;
 	Buddy* nb;
@@ -287,10 +263,10 @@ Nick* IRC::getNick(const im::Buddy& buddy) const
 	if(it == users.end())
 		return NULL;
 	else
-		return it->second;
+		return nb;
 }
 
-Nick* IRC::getNick(const im::Conversation& conv) const
+ConvNick* IRC::getNick(const im::Conversation& conv) const
 {
 	map<string, Nick*>::const_iterator it;
 	ConvNick* n;
@@ -302,7 +278,34 @@ Nick* IRC::getNick(const im::Conversation& conv) const
 	if(it == users.end())
 		return NULL;
 	else
-		return it->second;
+		return n;
+}
+
+vector<Nick*> IRC::matchNick(string pattern) const
+{
+	map<string, Nick*>::const_iterator it;
+	vector<Nick*> result;
+
+	if (pattern.find('!') == string::npos)
+	{
+		if (pattern.find('@') != string::npos)
+			pattern = "*!" + pattern;
+		else if (pattern.find(':') != string::npos)
+			pattern = "*!*@" + pattern;
+		else
+			pattern = pattern + "!*@*";
+	}
+	else if (pattern.find('@') == string::npos)
+		pattern += "@*";
+
+	for (it = users.begin(); it != users.end(); ++it)
+	{
+		string longname = it->second->getLongName();
+		if (!fnmatch(pattern.c_str(), longname.c_str(), FNM_NOESCAPE|FNM_CASEFOLD))
+			result.push_back(it->second);
+	}
+
+	return result;
 }
 
 void IRC::removeNick(string nickname)
@@ -322,6 +325,7 @@ void IRC::removeNick(string nickname)
 					(*dcc)->setPeer(NULL);
 				++dcc;
 			}
+		it->second->getServer()->removeNick(it->second);
 		delete it->second;
 		users.erase(it);
 	}
@@ -331,7 +335,10 @@ void IRC::cleanUpNicks()
 {
 	map<string, Nick*>::iterator it;
 	for(it = users.begin(); it != users.end(); ++it)
+	{
+		it->second->getServer()->removeNick(it->second);
 		delete it->second;
+	}
 	users.clear();
 }
 
@@ -407,14 +414,10 @@ void IRC::setMotd(const string& path)
 void IRC::quit(string reason)
 {
 	user->send(Message(MSG_ERROR).addArg("Closing Link: " + reason));
-
-	if(read_id >= 0)
-		g_source_remove(read_id);
-	read_id = -1;
-
 	user->close();
-	close(fd);
-	fd = -1;
+
+	delete sockw;
+	sockw = NULL;
 
 	poll->kill(this);
 }
@@ -425,20 +428,18 @@ void IRC::sendWelcome()
 	   user->getIdentname().empty())
 		return;
 
-	if(user->getPassword().empty())
-	{
-		quit("Please set a password");
-		return;
-	}
-
 	try
 	{
-		im_auth = im::Auth::build(this, user->getNickname());
-
-		if(!im_auth->exists())
+		im_auth = im::Auth::validate(this, user->getNickname(), user->getPassword());
+		if (!im_auth)
 		{
-			/* New user. */
+			if (im::IM::exists(user->getNickname()))
+			{
+				quit("Incorrect credentials");
+				return;
+			}
 
+			/* New User */
 			string global_passwd = conf.GetSection("irc")->GetItem("password")->String();
 			if(global_passwd != " " && user->getPassword() != global_passwd)
 			{
@@ -446,17 +447,15 @@ void IRC::sendWelcome()
 				return;
 			}
 
-			im_auth->create(user->getPassword());
-		}
-		else if(!im_auth->authenticate(user->getPassword()))
-		{
-			quit("Incorrect password");
-			return;
+			im_auth = im::Auth::generate(this, user->getNickname(), user->getPassword());
+			if (!im_auth)
+				quit("Creation of new account failed");
 		}
 
 		im = im_auth->getIM();
 
 		user->setFlag(Nick::REGISTERED);
+		poll->ipc_send(Message(MSG_USER).addArg(getUser()->getNickname()));
 
 		user->send(Message(RPL_WELCOME).setSender(this).setReceiver(user).addArg("Welcome to the Minbif IRC gateway, " + user->getNickname() + "!"));
 		user->send(Message(RPL_YOURHOST).setSender(this).setReceiver(user).addArg("Your host is " + getServerName() + ", running " MINBIF_VERSION));
@@ -465,10 +464,18 @@ void IRC::sendWelcome()
 		m_motd(Message());
 
 		im->restore();
+
+		if (im->isAway())
+		{
+			user->setAwayMessage("Away");
+			user->send(Message(RPL_NOWAWAY).setSender(this)
+					               .setReceiver(user)
+						       .addArg("You have been marked as being away"));
+		}
 	}
 	catch(im::IMError& e)
 	{
-		quit("Unable to initialize IM");
+		quit("Unable to initialize IM: " + e.Reason());
 	}
 }
 
@@ -492,77 +499,75 @@ bool IRC::ping(void*)
 
 void IRC::notice(Nick* nick, string msg)
 {
-	nick->send(Message(MSG_NOTICE).setSender(this)
-				      .setReceiver(user)
-				      .addArg(msg));
+	string tmp;
+	while((tmp = stringtok(msg, "\n\r")).empty() == false)
+		nick->send(Message(MSG_NOTICE).setSender(this)
+					      .setReceiver(user)
+					      .addArg(tmp));
 }
 
 void IRC::privmsg(Nick* nick, string msg)
 {
-	nick->send(Message(MSG_PRIVMSG).setSender(this)
-				       .setReceiver(user)
-				       .addArg(msg));
+	string tmp;
+	while((tmp = stringtok(msg, "\n\r")).empty() == false)
+		nick->send(Message(MSG_PRIVMSG).setSender(this)
+					       .setReceiver(user)
+					       .addArg(tmp));
 }
 
 bool IRC::readIO(void*)
 {
-	static char buf[1024];
-	string sbuf, line;
-	ssize_t r;
-
-	if((r = read(fd, buf, sizeof buf - 1)) <= 0)
+	try
 	{
-		if(r == 0)
-			this->quit("Connection reset by peer...");
-		else if(!sockerr_again())
-			this->quit(string("Read error: ") + strerror(errno));
-		else
-			return true; // continue...
-		return false;
-	}
-	buf[r] = 0;
-	sbuf = buf;
+		string sbuf, line;
 
-	while((line = stringtok(sbuf, "\r\n")).empty() == false)
-	{
-		Message m = Message::parse(line);
-		b_log[W_PARSE] << "<< " << line;
-		size_t i;
-		for(i = 0;
-		    i < (sizeof commands / sizeof *commands) &&
-		    strcmp(commands[i].cmd, m.getCommand().c_str());
-		    ++i)
-			;
+		sbuf = sockw->Read();
 
-		user->setLastReadNow();
-
-		if(i >= (sizeof commands / sizeof *commands))
-			user->send(Message(ERR_UNKNOWNCOMMAND).setSender(this)
-							   .setReceiver(user)
-							   .addArg(m.getCommand())
-							   .addArg("Unknown command"));
-		else if(m.countArgs() < commands[i].minargs)
-			user->send(Message(ERR_NEEDMOREPARAMS).setSender(this)
-							   .setReceiver(user)
-							   .addArg(m.getCommand())
-							   .addArg("Not enough parameters"));
-		else if(commands[i].flags && !user->hasFlag(commands[i].flags))
+		while((line = stringtok(sbuf, "\r\n")).empty() == false)
 		{
-			if(!user->hasFlag(Nick::REGISTERED))
-				user->send(Message(ERR_NOTREGISTERED).setSender(this)
-								     .setReceiver(user)
-								     .addArg("Register first"));
+			Message m = Message::parse(line);
+			b_log[W_PARSE] << "<< " << line;
+			size_t i;
+			for(i = 0;
+			    i < (sizeof commands / sizeof *commands) &&
+			    strcmp(commands[i].cmd, m.getCommand().c_str());
+			    ++i)
+				;
+
+			user->setLastReadNow();
+
+			if(i >= (sizeof commands / sizeof *commands))
+				user->send(Message(ERR_UNKNOWNCOMMAND).setSender(this)
+								   .setReceiver(user)
+								   .addArg(m.getCommand())
+								   .addArg("Unknown command"));
+			else if(m.countArgs() < commands[i].minargs)
+				user->send(Message(ERR_NEEDMOREPARAMS).setSender(this)
+								   .setReceiver(user)
+								   .addArg(m.getCommand())
+								   .addArg("Not enough parameters"));
+			else if(commands[i].flags && !user->hasFlag(commands[i].flags))
+			{
+				if(!user->hasFlag(Nick::REGISTERED))
+					user->send(Message(ERR_NOTREGISTERED).setSender(this)
+									     .setReceiver(user)
+									     .addArg("Register first"));
+				else
+					user->send(Message(ERR_NOPRIVILEGES).setSender(this)
+									    .setReceiver(user)
+									    .addArg("Permission Denied: Insufficient privileges"));
+			}
 			else
-				user->send(Message(ERR_NOPRIVILEGES).setSender(this)
-						                    .setReceiver(user)
-								    .addArg("Permission Denied: Insufficient privileges"));
-		}
-		else
-		{
-			commands[i].count++;
-			(this->*commands[i].func)(m);
+			{
+				commands[i].count++;
+				(this->*commands[i].func)(m);
+			}
 		}
 	}
+	catch (sock::SockError &e)
+	{
+		quit(e.Reason());
+	}
 
 	return true;
 }
@@ -657,6 +662,19 @@ void IRC::m_version(Message message)
 				       .addArg(MINBIF_BUILD));
 }
 
+/** INFO */
+void IRC::m_info(Message message)
+{
+	for(size_t i = 0; infotxt[i] != NULL; ++i)
+		user->send(Message(RPL_INFO).setSender(this)
+				            .setReceiver(user)
+					    .addArg(infotxt[i]));
+
+	user->send(Message(RPL_ENDOFINFO).setSender(this)
+			                 .setReceiver(user)
+					 .addArg("End of /INFO list."));
+}
+
 /** WHO */
 void IRC::m_who(Message message)
 {
@@ -747,12 +765,21 @@ void IRC::m_whois(Message message)
 					    .setReceiver(user)
 					    .addArg(n->getNickname())
 					    .addArg(n->getAwayMessage()));
+
+	string status = n->getStatusMessage();
+	if (!status.empty())
+		user->send(Message(RPL_WHOISACTUALLY).setSender(this)
+						     .setReceiver(user)
+						     .addArg(n->getNickname())
+						     .addArg(status));
+
 	if(n->hasFlag(Nick::OPER))
 		user->send(Message(RPL_WHOISOPERATOR).setSender(this)
 				                     .setReceiver(user)
 						     .addArg(n->getNickname())
 						     .addArg("is an IRC Operator"));
 
+
 	CacaImage icon = n->getIcon();
 	try
 	{
@@ -963,7 +990,7 @@ void IRC::m_stats(Message message)
 		{
 			string accid = message.countArgs() > 1 ? message.getArg(1) : "";
 			im::Account account = im->getAccount(accid);
-			if(!account.isValid())
+			if(!account.isValid() || !account.isConnected())
 				user->send(Message(ERR_NOSUCHCHANNEL).setSender(this)
 								     .setReceiver(user)
 								     .addArg(accid)
@@ -986,6 +1013,22 @@ void IRC::m_stats(Message message)
 								     .addArg(t2s(commands[i].count))
 								     .addArg("0"));
 			break;
+		case 'o':
+		{
+			vector<ConfigSection*> opers = conf.GetSection("irc")->GetSectionClones("oper");
+			for(vector<ConfigSection*>::iterator it = opers.begin(); it != opers.end(); ++it)
+			{
+				ConfigSection* oper = *it;
+
+				user->send(Message(RPL_STATSOLINE).setSender(this)
+						                  .setReceiver(user)
+								  .addArg("O")
+								  .addArg(oper->GetItem("email")->String())
+								  .addArg("*")
+								  .addArg(oper->GetItem("login")->String()));
+			}
+			break;
+		}
 		case 'p':
 		{
 			map<string, im::Protocol> m = im->getProtocolsList();
@@ -997,12 +1040,24 @@ void IRC::m_stats(Message message)
 			}
 			break;
 		}
+		case 'u':
+		{
+			unsigned now = time(NULL) - uptime;
+			gchar *msg = g_strdup_printf("Server Up %d days, %d:%02d:%02d", now / 86400, (now / 3600) % 24, (now / 60) % 60, now % 60);
+			user->send(Message(RPL_STATSUPTIME).setSender(this)
+			                                   .setReceiver(user)
+							   .addArg(msg));
+			g_free(msg);
+			break;
+		}
 		default:
 			arg = "*";
 			notice(user, "a (aways) - List all away messages availables");
 			notice(user, "c (chat params) - List all chat parameters for a specific account");
 			notice(user, "m (commands) - List all IRC commands");
+			notice(user, "o (opers) - List all opers accounts");
 			notice(user, "p (protocols) - List all protocols");
+			notice(user, "u (uptime) - Display the server uptime");
 			break;
 	}
 	user->send(Message(RPL_ENDOFSTATS).setSender(this)
@@ -1026,7 +1081,7 @@ void IRC::m_connect(Message message)
 		{
 			found = true;
 			account.connect();
-			Channel* chan = getChannel(account.getStatusChannel());
+			Channel* chan = account.getStatusChannel();
 			if(chan)
 				user->join(chan, ChanUser::OP);
 
@@ -1079,7 +1134,8 @@ void IRC::m_map(Message message)
 				message.rebuildWithQuotes();
 				if(message.countArgs() < 2)
 				{
-					notice(user, "Usage: /MAP add PROTO USERNAME PASSWD [OPTIONS] [CHANNEL]");
+					notice(user, "Usage: /MAP add PROTO USERNAME [OPTIONS]");
+					notice(user, "To display available options: /MAP add PROTO");
 					return;
 				}
 				string protoname = message.getArg(1);
@@ -1095,85 +1151,88 @@ void IRC::m_map(Message message)
 					return;
 				}
 
-				vector<im::Protocol::Option> options = proto.getOptions();
-				if(message.countArgs() < 4)
+				im::Protocol::Options options = proto.getOptions();
+				if(message.countArgs() < 3)
 				{
 					string s;
-					FOREACH(vector<im::Protocol::Option>, options, it)
+					FOREACH(im::Protocol::Options, options, it)
 					{
 						if(!s.empty()) s += " ";
 						s += "[-";
-						switch(it->getType())
+						switch(it->second.getType())
 						{
-							case PURPLE_PREF_BOOLEAN:
-								s += "[!]" + it->getName();
+							case im::Protocol::Option::SERVER_ALIASES:
+							case im::Protocol::Option::BOOL:
+								s += "[!]" + it->second.getName();
 								break;
-							case PURPLE_PREF_STRING:
-								s += it->getName() + " value";
+							case im::Protocol::Option::ACCID:
+							case im::Protocol::Option::STATUS_CHANNEL:
+							case im::Protocol::Option::PASSWORD:
+							case im::Protocol::Option::STR:
+								s += it->second.getName() + " value";
 								break;
-							case PURPLE_PREF_INT:
-								s += it->getName() + " int";
+							case im::Protocol::Option::INT:
+								s += it->second.getName() + " int";
 								break;
-							default:
+							case im::Protocol::Option::NONE:
 								break;
 						}
 						s += "]";
 					}
-					notice(user, "Usage: /MAP add " + proto.getID() + " USERNAME PASSWD " + s + " [CHANNEL]");
+					notice(user, "Usage: /MAP add " + proto.getID() + " USERNAME " + s);
 					return;
 				}
-				string username, password, channel;
-				for(size_t i = 2; i < message.countArgs(); ++i)
-				{
-					string s = message.getArg(i);
-					if(username.empty())
-						username = s;
-					else if(password.empty())
-						password = s;
-					else if(s[0] == '-')
+				string username;
+				try {
+					for(size_t i = 2; i < message.countArgs(); ++i)
 					{
-						size_t name_pos = 1;
-						string value = "true";
-						if(s[1] == '!')
+						string s = message.getArg(i);
+						if(username.empty())
+							username = s;
+						else if(s[0] == '-')
 						{
-							value = "false";
-							name_pos++;
-						}
+							size_t name_pos = 1;
+							string value = "true";
+							if(s[1] == '!')
+							{
+								value = "false";
+								name_pos++;
+							}
 
-						vector<im::Protocol::Option>::iterator it = std::find(options.begin(), options.end(), s.substr(name_pos));
-						if(it == options.end())
-						{
-							notice(user, "Error: Option '" + s + "' is unknown");
-							return;
+							im::Protocol::Options::iterator it = options.find(s.substr(name_pos));
+							if (it == options.end())
+							{
+								notice(user, "Error: Option '" + s + "' does not exist");
+								return;
+							}
+							switch(it->second.getType())
+							{
+								case im::Protocol::Option::BOOL:
+								case im::Protocol::Option::SERVER_ALIASES:
+									/* No input value needed, already got above */
+									break;
+								default:
+									if(i+1 < message.countArgs())
+										value = message.getArg(++i);
+									else
+									{
+										notice(user, "Error: Option '" + s + "' needs a value");
+										return;
+									}
+							}
+							it->second.setValue(value);
 						}
-						if(it->getType() == PURPLE_PREF_BOOLEAN)
-						{
-							/* No input value needed, already got above */
-						}
-						else if(i+1 < message.countArgs())
-							value = message.getArg(++i);
+						else if (options["password"].getValue().empty())
+							options["password"].setValue(s);
 						else
-						{
-							notice(user, "Error: Option '" + s + "' needs a value");
-							return;
-						}
-						it->setValue(value);
+							options["status_channel"].setValue(s);
 					}
-					else if(channel.empty())
-					{
-						channel = s;
-						if(!Channel::isStatusChannel(channel))
-						{
-							notice(user, "Error: Status channel must start with '&'");
-							return;
-						}
-					}
-				}
 
-				added_account = im->addAccount(proto, username, password, options, register_account);
-				if(channel.empty())
-					channel = "&minbif";
-				added_account.setStatusChannel(channel);
+					added_account = im->addAccount(proto, username, options, register_account);
+				} catch (im::Protocol::OptionError& e) {
+					notice(user, "Error: " + e.Reason());
+					return;
+				}
 
 				break;
 			}
@@ -1191,20 +1250,30 @@ void IRC::m_map(Message message)
 					notice(user, "Error: Account " + message.getArg(1) + " is unknown");
 					return;
 				}
-				vector<im::Protocol::Option> options = account.getOptions();
+				im::Protocol::Options options = account.getOptions();
 				if(message.countArgs() < 3)
 				{
 					notice(user, "-- Parameters of account " + account.getServername() + " --");
-					FOREACH(vector<im::Protocol::Option>, options, it)
+					FOREACH(im::Protocol::Options, options, it)
 					{
-						im::Protocol::Option& option = *it;
-						notice(user, option.getName() + " = " + option.getValue());
+						im::Protocol::Option& option = it->second;
+						string value;
+						switch(option.getType())
+						{
+							case im::Protocol::Option::PASSWORD:
+								value = "*******";
+								break;
+							default:
+								value = option.getValue();
+								break;
+						}
+						notice(user, option.getName() + " = " + value);
 					}
 					return;
 				}
-				vector<im::Protocol::Option>::iterator option;
+				im::Protocol::Options::iterator option;
 				for(option = options.begin();
-				    option != options.end() && option->getName() != message.getArg(2);
+				    option != options.end() && option->second.getName() != message.getArg(2);
 				    ++option)
 					;
 
@@ -1215,7 +1284,7 @@ void IRC::m_map(Message message)
 				}
 
 				if(message.countArgs() < 4)
-					notice(user, option->getName() + " = " + option->getValue());
+					notice(user, option->second.getName() + " = " + option->second.getValue());
 				else
 				{
 					string value;
@@ -1225,21 +1294,25 @@ void IRC::m_map(Message message)
 						value += message.getArg(i);
 					}
 
-					if(option->getType() == PURPLE_PREF_BOOLEAN && value != "true" && value != "false")
+					if(option->second.getType() == im::Protocol::Option::BOOL && value != "true" && value != "false")
 					{
-						notice(user, "Error: Option '" + option->getName() + "' is a boolean ('true' or 'false')");
+						notice(user, "Error: Option '" + option->second.getName() + "' is a boolean ('true' or 'false')");
 						return;
 					}
 					/* TODO check if value is an integer if option is an integer */
-					option->setValue(value);
-					if(option->getType() == PURPLE_PREF_INT)
-						notice(user, option->getName() + " = " + t2s(option->getValueInt()));
+					option->second.setValue(value);
+					if(option->second.getType() == im::Protocol::Option::INT)
+						notice(user, option->second.getName() + " = " + t2s(option->second.getValueInt()));
 					else
-						notice(user, option->getName() + " = " + option->getValue());
-					account.setOptions(options);
+						notice(user, option->second.getName() + " = " + option->second.getValue());
+
+					try {
+						account.setOptions(options);
+					} catch(im::Protocol::OptionError& e) {
+						notice(user, "Error: " + e.Reason());
+					}
 				}
 				return;
-				break;
 			}
 			case 'D':
 			case 'd':
@@ -1312,20 +1385,27 @@ void IRC::m_map(Message message)
 	    it != accounts.end(); ++it)
 	{
 		map<string, im::Account>::iterator next = it;
+		Server* server = getServer(it->second.getServername());
 		string name;
 
 		if(++next == accounts.end())
-			name = "`-";
+			name = "`";
 		else
-			name = "|-";
+			name = "|";
+
+		if(it->second == added_account)
+			name += "-+";
+		else if(it->second.isConnecting())
+			name += "-*";
+		else if(it->second.isConnected())
+			name += "-";
+		else
+			name += " ";
 
 		name += it->second.getServername();
-		if(it->second == added_account)
-			name += " (added)";
-		else if(it->second.isConnecting())
-			name += " (connecting)";
-		else if(!it->second.isConnected())
-			name += " (disconnected)";
+
+		if(it->second.isConnected())
+			name += "  [" + t2s(server->countOnlineNicks()) + "/" + t2s(server->countNicks()) + "]";
 
 		user->send(Message(RPL_MAP).setSender(this)
 					   .setReceiver(user)
@@ -1349,16 +1429,19 @@ void IRC::m_admin(Message message)
 		bool display;
 		SettingBase* setting;
 	} settings[] = {
-		{ "password",      true,  new SettingPassword(this, im) },
-		{ "typing_notice", true,  new SettingTypingNotice(this, im) },
-		{ "away_idle",     true,  new SettingAwayIdle(this, im) },
-		{ "log_level",     true,  new SettingLogLevel(this, im) },
-		{ "proxy",         true,  new SettingProxy(this, im) },
-		{ "proxy_host",    true,  new SettingProxyHost(this, im) },
-		{ "proxy_port",    true,  new SettingProxyPort(this, im) },
-		{ "proxy_user",    true,  new SettingProxyUsername(this, im) },
-		{ "proxy_pass",    true,  new SettingProxyPassword(this, im) },
-		{ "minbif",        false, new SettingMinbif(this, im) },
+		{ "password",                  true,  new SettingPassword(this, im) },
+		{ "typing_notice",             true,  new SettingTypingNotice(this, im) },
+		{ "accept_nobuddies_messages", true,  new SettingAcceptNoBuddiesMessages(this, im) },
+		{ "voiced_buddies",            true,  new SettingVoicedBuddies(this, im) },
+		{ "server_aliases",            true,  new SettingServerAliases(this, im) },
+		{ "away_idle",                 true,  new SettingAwayIdle(this, im) },
+		{ "log_level",                 true,  new SettingLogLevel(this, im) },
+		{ "proxy",                     true,  new SettingProxy(this, im) },
+		{ "proxy_host",                true,  new SettingProxyHost(this, im) },
+		{ "proxy_port",                true,  new SettingProxyPort(this, im) },
+		{ "proxy_user",                true,  new SettingProxyUsername(this, im) },
+		{ "proxy_pass",                true,  new SettingProxyPassword(this, im) },
+		{ "minbif",                    false, new SettingMinbif(this, im) },
 	};
 
 	if(message.countArgs() == 0)
@@ -1470,7 +1553,8 @@ void IRC::m_join(Message message)
 					continue;
 				}
 
-				if(!account.joinChat(convname, param))
+				/* purple_url_decode() returns a static buffer, no free needed. */
+				if(!account.joinChat(purple_url_decode(convname.c_str()), param))
 				{
 					user->send(Message(ERR_NOSUCHCHANNEL).setSender(this)
 									     .setReceiver(user)
@@ -1478,7 +1562,7 @@ void IRC::m_join(Message message)
 									     .addArg("No such channel"));
 				}
 				else
-					g_timeout_add(500, g_callback_delete, new CallBack<IRC>(this, &IRC::check_channel_join, g_strdup(channame.c_str())));
+					g_timeout_add(1000, g_callback_delete, new CallBack<IRC>(this, &IRC::check_channel_join, g_strdup(channame.c_str())));
 
 #if 0
 				chan = new ConversationChannel(this, conv);
@@ -1769,6 +1853,7 @@ void IRC::m_kill(Message message)
 void IRC::m_svsnick(Message message)
 {
 	Nick* n = getNick(message.getArg(0));
+	Nick* n2 = NULL;
 	if(!n)
 	{
 		user->send(Message(ERR_NOSUCHNICK).setSender(this)
@@ -1794,7 +1879,7 @@ void IRC::m_svsnick(Message message)
 		return;
 	}
 
-	if(getNick(message.getArg(1), true))
+	if((n2 = getNick(message.getArg(1))) && n2 != n)
 	{
 		user->send(Message(ERR_NICKNAMEINUSE).setSender(this)
 				                     .setReceiver(user)
